import { RevealStateManager } from './RevealStateManager.js';
import { config } from './config.js';
function closest(dom, selector) {
    var _a;
    const r = dom.closest(selector);
    if (r)
        return r;
    const host = (_a = findParentShadowRoot(dom)) === null || _a === void 0 ? void 0 : _a.host;
    if (!host)
        return null;
    return closest(host, selector);
}
function findParentShadowRoot(x) {
    if (!x)
        return null;
    if (x.parentNode && x.parentNode.nodeType === document.DOCUMENT_FRAGMENT_NODE)
        return x.parentNode;
    return findParentShadowRoot(x.parentNode);
}
class ServerSideHTMLElement {
    constructor() {
        this.root = {
            innerHTML: '',
            querySelector: (_x) => null,
        };
        this.attachShadow = () => null;
    }
}
const PatchedHTMLElement = typeof globalThis.Window === 'undefined' ? ServerSideHTMLElement : HTMLElement;
export class AxRevealProvider extends PatchedHTMLElement {
    constructor() {
        super(...arguments);
        this.stateManager = new RevealStateManager();
    }
}
AxRevealProvider.ElementName = 'ax-reveal-provider';
export class AxRevealBoundary extends PatchedHTMLElement {
    constructor() {
        super();
        this.root = this.attachShadow({ mode: 'open' });
        /**
         * Update the position of your pointer.
         * @param ev Pointer event from the listener.
         */
        this.updatePointerPosition = (ev) => {
            this.waitForStorage((storage) => {
                storage.clientX = ev.clientX;
                storage.clientY = ev.clientY;
            });
        };
        this.handlePointerEnter = () => this.waitForStorage((storage) => storage.onPointerEnterBoundary());
        this.handlePointerLeave = () => this.waitForStorage((storage) => storage.onPointerLeaveBoundary());
        this.handlePointerMove = this.updatePointerPosition;
        this.handlePointerUp = (ev) => {
            if (ev.pointerType === 'mouse') {
                this.waitForStorage((storage) => storage.switchAnimation());
            }
            else {
                this.waitForStorage((storage) => storage.clearAll(false));
            }
        };
        this.handlePointerDown = (ev) => this.waitForStorage((storage) => {
            this.updatePointerPosition(ev);
            storage.initializeAnimation();
        });
        this.root.innerHTML = `
<div>
    <slot></slot>
</div>`;
    }
    get storage() {
        return this._storage;
    }
    set storage(newStorage) {
        const oldStorage = this._storage;
        if (oldStorage) {
            this.dispatchEvent(new CustomEvent(AxRevealBoundary.removeStorageEvent, { detail: oldStorage }));
            AxRevealBoundary.stateManager.removeBoundary(oldStorage);
        }
        if (newStorage) {
            this._storage = newStorage;
            this.dispatchEvent(new CustomEvent(AxRevealBoundary.attachStorageEvent, { detail: newStorage }));
        }
        if (oldStorage && newStorage) {
            this.dispatchEvent(new CustomEvent(AxRevealBoundary.replaceStorageEvent, {
                detail: { old: oldStorage, new: newStorage },
            }));
        }
    }
    waitForStorage(f) {
        if (this.storage === undefined) {
            this.addEventListener(AxRevealBoundary.attachStorageEvent, () => f(this.storage), { once: true });
        }
        else {
            f(this.storage);
        }
    }
    appendStorage(force = false) {
        if (!force && this.storage) {
            return;
        }
        const parent = closest(this, AxRevealProvider.ElementName);
        const stateManager = parent ? parent.stateManager : AxRevealBoundary.stateManager;
        this.storage = stateManager.newBoundary(this);
    }
    connectedCallback() {
        this.appendStorage(true);
        if (config.borderDetectionMode === 'strictEdge') {
            this.addEventListener('pointerenter', this.handlePointerEnter);
            this.addEventListener('pointerleave', this.handlePointerLeave);
            this.addEventListener('pointermove', this.handlePointerMove);
        }
        this.addEventListener('pointerdown', this.handlePointerDown);
        this.addEventListener('pointerup', this.handlePointerUp);
    }
    disconnectedCallback() {
        this.storage = undefined;
    }
}
AxRevealBoundary.ElementName = 'ax-reveal-bound';
AxRevealBoundary.removeStorageEvent = 'removeStorage';
AxRevealBoundary.attachStorageEvent = 'attachStorage';
AxRevealBoundary.replaceStorageEvent = 'replaceStorage';
AxRevealBoundary.stateManager = new RevealStateManager();
export class AxReveal extends PatchedHTMLElement {
    constructor() {
        super();
        this.root = this.attachShadow({ mode: 'open' });
        this.handlePointerDown = (ev) => {
            this.setPointerCapture(ev.pointerId);
        };
        this.root.innerHTML = `
<div class="ax-reveal">
    <canvas></canvas>
    <div class="content"><slot></slot></div>
</div>
<style>
    .ax-reveal { display: content; }
    .ax-reveal * { user-drag: none; -webkit-user-drag: none; touch-action: none; user-select: none; }
    .content { position: relative; }
    canvas { top: 0; left: 0; pointer-events: none; width: 100%; height: 100%; position: absolute; }
    :host { display: inline-block; position: relative; }
    :host([block]) { display: block; }
    :host([inline-block]) { display: inline-block; }
    :host([flex]) { display: flex; }
    :host([inline-flex]) { display: inline-flex; }
    :host([grid]) { display: grid; }
    :host([inline-grid]) { display: inline-grid; }
    ::slotted(*) { user-drag: none; -webkit-user-drag: none; touch-action: none; user-select: none; }
    ::slotted(button) { outline:none; }
</style>`;
        this.canvas = this.root.querySelector('canvas');
    }
    adoptedCallback() {
        this.disconnectedCallback();
        this.connectedCallback();
    }
    disconnectedCallback() {
        this.boundary && this.boundary.waitForStorage((storage) => storage.removeReveal(this.canvas));
    }
    connectedCallback() {
        this.boundary = closest(this, AxRevealBoundary.ElementName);
        if (!this.boundary)
            throw new SyntaxError('You must use ' + AxRevealBoundary.ElementName + ' as the boundary of reveal highlight!');
        this.boundary.waitForStorage((storage) => setTimeout(() => storage.addReveal(this.canvas, this), 0));
        this.addEventListener('pointerdown', this.handlePointerDown);
    }
}
AxReveal.ElementName = 'ax-reveal';
export class AxRevealNg extends PatchedHTMLElement {
    constructor() {
        super();
        this.root = this.attachShadow({ mode: 'open' });
        this.handlePointerDown = (ev) => {
            this.setPointerCapture(ev.pointerId);
        };
        this.root.innerHTML = `
<div class="ax-reveal">
    <svg>
        <defs>
            <radialGradient id="borderGrad" gradientUnits="userSpaceOnUse" r="10">
                <stop id="borderCenter" offset="0%" />
                <stop id="borderOut" offset="100%" />
            </radialGradient>
            <radialGradient id="fillGrad" gradientUnits="userSpaceOnUse" r="10">
                <stop id="fillCenter" offset="0%" />
                <stop id="fillOut" offset="100%" />
            </radialGradient>
            <radialGradient id="rippleGrad" gradientUnits="userSpaceOnUse" r="10">
                <stop id="rippleCenter" offset="0%" />
                <stop id="rippleMiddle" offset="0%" />
                <stop id="rippleOut" offset="0%" />
            </radialGradient>
        </defs>
        <path id="borderPath" fill="url(#borderGrad)" />
        <path id="fillPath" fill="url(#fillGrad)" />
        <path id="ripplePath" fill="url(#rippleGrad)" />
    </svg>
    <div class="content"><slot></slot></div>
</div>
<style>
    .ax-reveal { display: contents; }
    .ax-reveal * { user-drag: none; -webkit-user-drag: none; touch-action: none; user-select: none; }
    .content { position: relative; }
    svg { top: 0; left: 0; pointer-events: none; width: 100%; height: 100%; position: absolute; }
    :host { display: inline-block; position: relative; }
    :host([block]) { display: block; }
    :host([inline-block]) { display: inline-block; }
    :host([flex]) { display: flex; }
    :host([inline-flex]) { display: inline-flex; }
    :host([grid]) { display: grid; }
    :host([inline-grid]) { display: inline-grid; }
    ::slotted(*) { user-drag: none; -webkit-user-drag: none; touch-action: none; user-select: none; }
    ::slotted(button) { outline:none; }
</style>`;
        this.svg = this.root.querySelector('svg');
    }
    adoptedCallback() {
        this.disconnectedCallback();
        this.connectedCallback();
    }
    disconnectedCallback() {
        this.boundary && this.boundary.waitForStorage((storage) => storage.removeReveal(this.svg));
    }
    connectedCallback() {
        this.boundary = closest(this, AxRevealBoundary.ElementName);
        if (!this.boundary)
            throw new SyntaxError('You must use ' + AxRevealBoundary.ElementName + ' as the boundary of reveal highlight!');
        this.boundary.waitForStorage((storage) => setTimeout(() => storage.addReveal(this.svg, this), 0));
        this.addEventListener('pointerdown', this.handlePointerDown);
    }
}
AxRevealNg.ElementName = 'ax-reveal-ng';
//# sourceMappingURL=CustomElements.js.map